# サービス

<a id="preparing_to_use_the_service"></a>
## サービスを利用するには

サービスのインストールと設定の手順は、以下のとおりです。

1.  GridDBサーバパッケージ、クライアントパッケージのインストール
2.  クラスタを構成するすべてのノードの設定
3.  起動設定ファイルの設定

サービスで用いるファイルの種類です。

| 種類               | 意味                                          |
|--------------------|----------------------------------------------|
| systemd ユニットファイル | systemdのユニット定義ファイル。GridDBのサーバパッケージにより、/usr/lib/systemd/system/gridstore.service にインストールされ、GridDBサービスとしてシステムに登録される。 |
| サービススクリプト | OS起動時に自動的に実行されるスクリプトファイル。<br>GridDBのサーバパッケージにより、/usr/griddb/bin/gridstoreにインストールされる。 |
| PIDファイル        | gsserverプロセスのプロセスID(PID)のみを記載したファイル。gsserverプロセス起動時に、$GS_HOME/conf/gridstore.pidに作成される。     |
| 起動設定ファイル   | サービスの中で設定可能な変数を記載するファイル。<br>GridDBのサーバパッケージにより、/etc/sysconfig/gridstore/gridstore.confにインストールされる。 |


## パラメータ設定

GridDBのサービスの動作を制御するパラメータを用意しています。パラメータの一覧は以下のとおりです。

| パラメータ    | デフォルト                           | 説明                           |
|--------------|-------------------------------------|--------------------------------|
| GS_USER      | admin                               | GridDBのユーザ名               |
| GS_PASSWORD  | admin                               | `GS_USER` のパスワード         |
| CLUSTER_NAME | myCluster        | 参加するクラスタ名              |
| MIN_NODE_NUM  | 1                                  | 参加するクラスタの構成ノード数   |

パラメータを変更するには起動設定ファイル( `/etc/sysconfig/gridstore/gridstore.conf` )を編集します。

サーバパッケージのアップデートインストールやアンインストールのとき、起動設定ファイルは上書き・アンインストールされません。

【注意】
-   サービススクリプトに記載されているパラメータは直接編集しないでください。アンインストールやアップデートインストールの際に編集したファイルは失われます。 パラメータを変更するときは、起動設定ファイルを編集してください。

## ログ

サービスのログに関しては、ブートログ( `/var/log/boot.log` )や運用コマンドのログ( `$GS_HOME/log` )を参照してください。


## コマンド

GridDBのサービスのコマンドを以下に説明します。

### start

動作：

-   ノードを起動し、クラスタへ参加します。

``` example
$ sudo systemctl start gridstore
```

-   ノードの起動はgs_startnodeコマンド、クラスタへの参加はgs_joinclusterコマンドをそれぞれ用います。
-   gs_startnodeコマンド実行時、リカバリ処理の終了を待合せます。
-   gs_joinclusterコマンド実行時、クラスタ稼働まで待合せません。
-   `CLUSTER_NAME` にクラスタ名を設定します。
-   `MIN_NODE_NUM` にクラスタを構成するノード台数を設定します。

【注意】
-   クラスタを稼働する途中にエラーが発生した場合は、プロセスの終了処理が行われます。

### stop

動作：

-   クラスタから離脱してからノードを停止します。

``` example
$ sudo systemctl stop gridstore
```

-   プロセスが無くなれば完了、タイムアウト時間が経過すればエラーとします(終了コード150)。
-   サービスで起動したプロセスが無ければ、終了コード0とします。
-   クラスタからの離脱はgs_leaveclusterコマンドを用います。
-   ノードの停止前にgs_leaveclusterコマンドを実行します。
-   gs_leaveclusterコマンド実行時、クラスタ離脱を待ち合わせます。
-   gs_leaveclusterの終了コードにかかわらず、ノード停止処理を行います。

【注意】
-   **クラスタ全体を停止する時は、必ずgs_stopclusterコマンドを実行してから、サービスのstopで各ノードを離脱・停止してください。** gs_stopclusterコマンドでクラスタを停止しなかった場合、ノードの離脱のたびにデータ再配置が行われる可能性があります。データ再配置が頻繁に発生すると、ネットワークやディスクI/Oに負荷がかかる場合があります。クラスタを停止してからノードを離脱した場合はデータ再配置は行われません。不要なデータ再配置を防ぐために、必ずクラスタを停止してください。クラスタの停止は、運用コマンドgs_stopclusterや、統合運用管理gs_admin、gs_shなどを用いて実行してください。
-   運用コマンドやコマンド・インタプリタ(gs_sh)で起動したノードはサービスのstopで停止することはできません。 それぞれのツールで停止を行ってください。

### status

動作：

-   ノードのプロセスが実行中かどうかを表示します。

``` example
$ sudo systemctl status gridstore
```

### restart

動作：

-   stopとstartを連続で行います。

### condrestart

動作：

-   ロックファイルがあればrestartを行います。


## エラーメッセージ一覧

サービスのエラーメッセージは以下のとおりです。

| コード | メッセージ                   | 意味                                  |
|--------|-----------------------------|--------------------------------------|
| F00003 | Json load error             | 定義ファイルの読込みに失敗しました。     |
| F01001 | Stop service timed out      | ノード停止処理がタイムアウトしました。   |
| F01002 | Startnode error             | ノード起動処理でエラーが発生しました。   |
| F01003 | Startnode timed out         | ノード起動処理がタイムアウトしました。   |
| F01004 | Joincluster error           | クラスタ参加処理でエラーが発生しました。 |
| F01005 | Joincluster timed out       | クラスタ参加処理がタイムアウトしました。 |
| F01006 | Leavecluster error          | クラスタ離脱処理でエラーが発生しました。 |
| F02001 | Command execution error     | コマンド実行でエラーが発生しました。     |
| F02002 | Command execution timed out | コマンド実行がタイムアウトしました。     |

【メモ】
-   各コマンド実行でエラーが発生した場合、運用コマンドのエラーが合わせて表示・記録されます。 エラーへの対応の際は、運用コマンドの項(gs_startnode、gs_joincluster、gs_leavecluster)も合わせて参照してください。
